(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: ClipVolume.cpp
//
// Desc: Sample code showing how to use a vertex shader and a pixel shader to do
//       a non-planar clipping effect in D3D.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit ClipVolumeUnit;

interface

uses
  Windows, Messages, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders

//-----------------------------------------------------------------------------
// Name: struct MIRRORVERTEX
// Desc: Custom mirror vertex type
//-----------------------------------------------------------------------------
type
  PMirrorVertex = ^TMirrorVertex;
  TMirrorVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    color: DWORD;
  end;
  PAMirrorVertex = ^AMirrorVertex;
  AMirrorVertex = array[0..4] of TMirrorVertex;

const
  TMirrorVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_DIFFUSE);

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
    m_pFont: CD3DFont;        // Font for drawing text
    m_pFontSmall: CD3DFont;   // Font for drawing text
    m_ArcBall: CD3DArcBall;   // Mouse rotation utility

    m_pTeapotMesh: ID3DXMesh; // Teapot mesh
    m_pSphereMesh: ID3DXMesh; // Sphere mesh
    m_pEffect: ID3DXEffect;   // Effect containing the shaders

    m_matTeapotWorld: TD3DXMatrix;
    m_matSphereWorld: TD3DXMatrix;
    m_matArcballWorld: TD3DXMatrix;
    m_matProjection: TD3DXMatrix;

    m_fSphereMove: Single;      // Sphere per-second movement delta
    m_vecSphereCenter: TD3DXVector4; // Center of sphere

    m_bShowHelp: Boolean;
  protected
    function OneTimeSceneInit: HRESULT; override;
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function FinalCleanup: HRESULT; override;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

implementation

uses Resource;


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;
  m_strWindowTitle  := 'ClipVolume';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bShowCursorWhenFullscreen      := True;

  m_pFont           := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall      := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);
  m_ArcBall         := CD3DArcBall.Create;

  m_pTeapotMesh     := nil;
  m_pSphereMesh     := nil;
  m_pEffect         := nil;

  m_bShowHelp       := False;
end;


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HRESULT;
begin
  // Translation matrix for teapot mesh
  // NOTE: This translation is fixed and will preceded by the ArcBall
  //       rotation that is calculated per-frame in FrameMove
  D3DXMatrixTranslation(m_matTeapotWorld, 0.0, 0.0, 5.0);
  D3DXMatrixTranspose(m_matTeapotWorld, m_matTeapotWorld);

  // Translation matrix for sphere mesh
  // NOTE: This is built per-frame in FrameMove

  // VIEW matrix for the entire scene
  // NOTE: This is fixed to an identity matrix

  // PROJECTION matrix for the entire scene
  // NOTE: The projection is based on the window dimensions
  //       and is built in RestoreDeviceObjects

  // Initial per-second movement delta for sphere
  m_fSphereMove := 0.6;

  // Initial location of center of sphere
  // NOTE: .xyz = center of sphere
  //       .w   = radius of sphere
  m_vecSphereCenter := D3DXVector4(0.0, 0.0, 5.0, 1.0);

  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LONG_PTR(LoadCursor(0, IDC_SIZEALL)));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
begin
  // Update translation matrix for sphere
  if (m_vecSphereCenter.x > 2.0) then m_fSphereMove := -0.6 else
  if (m_vecSphereCenter.x < -2.0) then m_fSphereMove := 0.6;

  m_vecSphereCenter.x := m_vecSphereCenter.x + m_fSphereMove * m_fElapsedTime;
  D3DXMatrixTranslation(m_matSphereWorld, m_vecSphereCenter.x,
                                          m_vecSphereCenter.y,
                                          m_vecSphereCenter.z);
  D3DXMatrixTranspose(m_matSphereWorld, m_matSphereWorld);

  // Update rotation matrix from ArcBall
  D3DXMatrixTranspose(m_matArcballWorld, m_ArcBall.GetRotationMatrix^);

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
var
  numPasses: LongWord;
  iPass: Integer;
  hTechniqueTeapot: TD3DXHandle;
  hTechniqueSphere: TD3DXHandle;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     D3DCOLOR_RGBA(0,0,0,0), 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Draw teapot mesh using teapot effect
    m_pd3dDevice.SetVertexShaderConstantF( 0, @m_matArcballWorld, 4);
    m_pd3dDevice.SetVertexShaderConstantF( 4, @m_matTeapotWorld,  4);
    m_pd3dDevice.SetVertexShaderConstantF( 8, @m_matProjection,   4);
    m_pd3dDevice.SetVertexShaderConstantF(12, @m_vecSphereCenter, 1);
    hTechniqueTeapot := m_pEffect.GetTechniqueByName('Teapot');
    m_pEffect.SetTechnique(hTechniqueTeapot);
    m_pEffect._Begin(@numPasses, 0);
    for iPass := 0 to numPasses - 1 do
    begin
      m_pEffect.Pass(iPass);
      m_pTeapotMesh.DrawSubset(0);
    end;
    m_pEffect._End;

    // Draw sphere mesh using sphere effect
    m_pd3dDevice.SetVertexShaderConstantF(4, @m_matSphereWorld, 4);
    hTechniqueSphere := m_pEffect.GetTechniqueByName('Sphere');
    m_pEffect.SetTechnique(hTechniqueSphere);
    m_pEffect._Begin(@numPasses, 0);
    for iPass := 0 to numPasses - 1 do
    begin
      m_pEffect.Pass(iPass);
      m_pSphereMesh.DrawSubset(0);
    end;
    m_pEffect._End;

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // Toggle help text
    if m_bShowHelp then
    begin
      m_pFontSmall.DrawText(2, 42, D3DCOLOR_ARGB(255,0,255,255),
                            'Use mouse to rotate the teapot');
    end else
    begin
        m_pFontSmall.DrawText(2, 42, D3DCOLOR_ARGB(255,255,0,255),
                              'Press F1 for Help');
    end;

    // End the scene
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects:HRESULT;
var
  strPath: array[0..MAX_PATH-1] of Char;
  dwShaderFlags: DWORD;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  // Create teapot mesh and sphere mesh
  Result:= D3DXCreateTeapot(m_pd3dDevice, m_pTeapotMesh, nil);
  if FAILED(Result) then Exit;

  Result:= D3DXCreateSphere(m_pd3dDevice, 1.0, 30, 30, m_pSphereMesh, nil);
  if FAILED(Result) then Exit;

  // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
  // Debugging vertex shaders requires either REF or software vertex processing, and debugging
  // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
  // experience in the shader debugger.  It enables source level debugging, prevents instruction
  // reordering, prevents dead code elimination, and forces the compiler to compile against the next
  // higher available software target, which ensures that the unoptimized shaders do not exceed
  // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
  // will be unoptimized and forced into software.  See the DirectX documentation for more information
  // about using the shader debugger.
  dwShaderFlags := 0;
  {$IFDEF DEBUG_VS}
     dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
     dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
  {$ENDIF}

  // Load effect
  Result:= DXUtil_FindMediaFileCch(strPath, MAX_PATH, 'ClipVolume.fx');
  if FAILED(Result) then Exit;

  Result:= D3DXCreateEffectFromFile(m_pd3dDevice, strPath, nil, nil,
                                    dwShaderFlags, nil, m_pEffect, nil);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  fAspect: Single;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height);

  // Update projection matrix based on window dimensions
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(m_matProjection, D3DX_PI / 4, fAspect, 1.0, 60.0);
  D3DXMatrixTranspose(m_matProjection, m_matProjection);

  if (m_pEffect <> nil) then m_pEffect.OnResetDevice;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  if (m_pEffect <> nil) then m_pEffect.OnLostDevice;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  SAFE_RELEASE(m_pTeapotMesh);
  SAFE_RELEASE(m_pSphereMesh);
  SAFE_RELEASE(m_pEffect);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
       ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    Result:= E_FAIL;
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) then Exit;
  {$ENDIF}

  // Need to support post-pixel processing (for alpha blending)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
       D3DRTYPE_SURFACE, backBufferFormat))
  then Exit;

  // Device should support at least both VS.1.1 and PS.1.1
  if (pCaps.VertexShaderVersion >= D3DVS_VERSION(1, 1)) and
     (pCaps.PixelShaderVersion  >= D3DPS_VERSION(1, 1)) then
  begin
    Result:= S_OK;
    Exit;
  end;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle user input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall.HandleMessages(hWnd, uMsg, wParam, lParam);

  if (uMsg = WM_COMMAND) then
  begin
    case LOWORD(wParam) of
      IDM_TOGGLEHELP: m_bShowHelp := not m_bShowHelp;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

